{ ------------------------------------------------------------------------- }
{ Disciplina: Inteligencia Artificial                                       }
{ Prof.: Ines Dutra                                                         }
{ Alunas: Aline Marins Paes                                                 }
{         Paula Fernanda M. V. de Carvalho                                  }
{ ------------------------------------------------------------------------- }

program jogo_dos_8;
uses crt;

{ -------------------------- estruturas globais --------------------------- }
type
   op = array[0..3] of string[5];
   st = array[0..2,0..2] of integer;

   {estrutura de representacao do problema}
   problem = record
      operators     : op;
      initial_state : st;
      goal_state    : st;
   end;

   {nos que representam cada estado gerado pela busca}
   node = ^element;
   element = record
      state  : st;            {estado do no                        }
      depth  : integer;       {profundidade do no                  }
      cost   : integer;       {custo do no                         }
      heur   : integer;       {valor da heuristica do no           }
      oper   : string[5];     {operador utilizado para gerar o no  }
      father : node;          {ponteiro para o no pai              }
      next   : node;          {ponteiro para o proximo no da lista }
   end;

   {mantem os ponteiros para a lista}
   list = record
      before : node;
      start  : node;
      final  : node;
   end;

var
   jogo_dos_oito : problem;
   problem_node  : node;
   nodes_created : integer;
   i             : integer;

{ imprime no ---------------------------------------------------------------}
procedure print_node(pnode : node);
var
   i, l, c : integer;

begin
   if pnode<>nil then
   begin
      write(' depth: ', pnode^.depth);
      write(', cost: ', pnode^.cost);
      write(', heur: ', pnode^.heur);
      write(', oper: ');
      if (pnode^.oper<>'') then writeln(pnode^.oper) else writeln(' - ');

      for l:=0 to 2 do
      begin
         for c:=0 to 2 do
         begin
            if (pnode^.state[l,c] = 9) then  write ('  ')
            else write(pnode^.state[l,c], ' ');
         end;
         writeln;
      end;
   end;
end;

{ imprime lista ------------------------------------------------------------}
procedure print_list(plist : list);
var
   aux : node;

begin
   aux := plist.start;

   while aux<>nil do
   begin
      print_node(aux);
      writeln;
      aux:=aux^.next;
   end;
end;

{ copia estados ------------------------------------------------------------}
procedure Copy_state(orig:st; var dest:st);
var
   l, c : integer;

begin
   for l:=0 to 2 do
      for c:=0 to 2 do
         dest[l,c]:=orig[l,c];
end;

{ heuristica que sera utilizada nos metodos informados de busca------------ }
function Heuristic(state:st):integer;
var
   l, c, h : integer;

begin
   h:=0;

   for l:=0 to 2 do
      for c:=0 to 2 do
         if (state[l,c]<>jogo_dos_oito.goal_state[l,c]) then h:=h+1;

   Heuristic:=h;
end;

{ cria cada no da lista --------------------------------------------------- }
function Make_node(pfather:node; state_node:st; poper:string):node;
var
   l, c : integer;
   new_node : node;

begin
   new(new_node);

   {estado}
   Copy_state(state_node, new_node^.state);

   {profundidade e custo}
   if pfather<>nil then
   begin
      new_node^.depth := pfather^.depth +1;
      new_node^.cost := pfather^.cost +1;
   end
   else begin
      new_node^.depth := 0;
      new_node^.cost := 0;
   end;

   {heuristica para busca best-first}
   new_node^.heur := Heuristic(state_node);

   {operador utilizado para gerar o no}
   new_node^.oper := poper;

   {ponteiro para o pai}
   new_node^.father := pfather;

   {ponteiro para o proximo da lista}
   new_node^.next := nil;

   nodes_created:=nodes_created+1; {incrementa num de nos alocados}

   Make_node := new_node;
end;

{ cria a lista usada no problema, e' chamada uma unica vez ---------------- }
procedure Make_list(var plist:list; pnode:node);
begin
   plist.start:=pnode;
   plist.final:=pnode;
end;

{ desenfileira um no da lista (sempre o da primeira posicao) -------------- }
function Remove_front(var plist:list):node;
var
   temp : node;

begin
   temp:=plist.start;
   plist.before:=temp;
   plist.start:=plist.start^.next;

   if (plist.start=nil) then plist.final:=nil; {se lista vazia}

   Remove_front:=temp;
end;

{ verifica se a lista esta vazia ------------------------------------------ }
function IsEmpty(plist:list):boolean;
begin
   if (plist.start=nil) then IsEmpty:=true else IsEmpty:=false;
end;

{ enfileira o no na lista de acordo com o tipo passado -------------------- }
procedure Enqueue(var pnode:node; var plist:list; enqueue_fn:string);
var
   aux, bef : node;

begin
   if IsEmpty(plist) then
   begin
      if (plist.before<>nil) then plist.before^.next:=pnode;
      plist.start:=pnode;
      plist.final:=pnode;
   end
   else begin
      if (enqueue_fn = 'ENQUEUE_AT_END') then {busca em largura}
      begin
         plist.final^.next:=pnode;
         plist.final:=pnode;
      end
      else if (enqueue_fn = 'ENQUEUE_AT_FRONT') then {busca em profundidade}
      begin
         if (plist.before<>nil) then plist.before^.next:=pnode;
         pnode^.next:=plist.start;
         plist.start:=pnode;
      end
      else if (enqueue_fn = 'ENQUEUE_AC_TO_COST') then {busca custo uniforme}
      begin
         aux:=plist.start;
         while ((aux^.cost <= pnode^.cost) and (aux<>nil)) do
         begin
            bef:=aux;
            aux:=aux^.next;
         end;
      end
      else if (enqueue_fn = 'ENQUEUE_AC_TO_HEUR') then {busca gulosa}
      begin
         aux:=plist.start;
         while ((aux^.heur <= pnode^.heur) and (aux<>nil)) do
         begin
            bef:=aux;
            aux:=aux^.next;
         end;
      end
      else if (enqueue_fn = 'ENQUEUE_AC_TO_HEUR+COST') then {busca A*}
      begin
         aux:=plist.start;
         while (((aux^.heur + aux^.cost) <= (pnode^.heur + pnode^.cost))
                and (aux <> nil)) do
         begin
            bef:=aux;
            aux:=aux^.next;
         end;
      end;
      if ((enqueue_fn = 'ENQUEUE_AC_TO_COST') or
            (enqueue_fn = 'ENQUEUE_AC_TO_HEUR') or
            (enqueue_fn = 'ENQUEUE_AC_TO_HEUR+COST')) then
      begin
         if (aux=plist.start) then
         begin
            if (plist.before<>nil) then plist.before^.next:=pnode;
            pnode^.next:=plist.start;
            plist.start:=pnode;
         end
         else if (aux=nil) then
         begin
            plist.final^.next:=pnode;
            plist.final:=pnode;
         end
         else begin
            bef^.next:=pnode;
            pnode^.next:=aux;
         end;
      end;
   end;
end;

{ testa se o estado alcancado corresponde ao objetivo --------------------- }
function Goal_test(state:st; goal:st):boolean;
var
   l, c : integer;
   flag : boolean;

begin
   flag:=true;

   for l:=0 to 2 do
      for c:=0 to 2 do
         if (state[l,c]<>goal[l,c]) then flag:=false;

   Goal_test:=flag;
end;

{ percorre a lista para exibir ao usuario a solucao do problema ----------- }
function Path(pnode:node):node;
var
   i, l, c : integer;
   nodes_path : array[0..999] of node;
   temp : node;

begin
   temp:=pnode;

   for i:=0 to 99 do
      nodes_path[i]:=nil;

   i:=0;

   while (temp^.father<>nil) do
   begin
      nodes_path[i]:=temp;
      temp:=temp^.father;
      i:=i+1;
   end;
   nodes_path[i]:=temp;

   while (i>=0) do
   begin
      print_node(nodes_path[i]);
      i:=i-1;
   end;

   Path:=temp;
end;

{ desaloca todos os nos gerados ------------------------------------------- }
procedure Destroy(var pnode:node);
var
   temp : node;

begin
   temp:=pnode;

   while (temp<>nil) do
   begin
      pnode:=pnode^.next;
      dispose(temp);
      temp:=pnode;
   end;
end;

{ desaloca nos nas buscas BPL e BPI --------------------------------------- }
procedure Destroy_node(var pnode:node; var plist:list);
var
   next, father, temp : node;

begin
   next:=pnode^.next;
   father:=pnode^.father;

   if IsEmpty(plist) then
   begin
      dispose(plist.before);
      plist.before:=nil;
   end
   else begin
      if (pnode^.father<>next^.father) then
      begin
         while (father^.next<>next) do
         begin
            temp:=father^.next;

            father^.next:=temp^.next;
            dispose(temp);

            if (next^.next<>nil) then nodes_created:=nodes_created-1;
         end;

         pnode:=nil;

         plist.start:=next;
         if (plist.start=nil) then plist.final:=nil;
      end;
   end;
end;

{ expande nos de acordo com operadores ------------------------------------ }
procedure Expand(pnode:node; var plist:list; enqueue_fn:string);
type
   bl = record
      l : integer;
      c : integer;
   end;

var
   i, l, c  : integer;
   aux      : st;
   blank    : bl;
   new_node : node;

begin
   {verifica onde esta o quadrado vazio}
   for l:=0 to 2 do
   begin
      for c:=0 to 2 do
      begin
         if (pnode^.state[l,c]=9) then begin
            blank.l:=l;
            blank.c:=c;
         end;
      end;
   end;

   for i:=0 to 3 do
   begin
      case i of
      0: if (blank.c-1>=0) then begin {LEFT}
            Copy_state(pnode^.state, aux);
            aux[blank.l,blank.c]:=aux[blank.l,blank.c-1];
            aux[blank.l,blank.c-1]:=9;
            new_node := Make_node(pnode, aux, 'LEFT');
            Enqueue(new_node, plist, enqueue_fn);
         end;
      1: if (blank.c+1<=2) then begin {RIGHT}
            Copy_state(pnode^.state, aux);
            aux[blank.l,blank.c]:=aux[blank.l,blank.c+1];
            aux[blank.l,blank.c+1]:=9;
            new_node := Make_node(pnode, aux, 'RIGHT');
            Enqueue(new_node, plist, enqueue_fn);
         end;
      2: if (blank.l-1>=0) then begin {UP}
            Copy_state(pnode^.state, aux);
            aux[blank.l,blank.c]:=aux[blank.l-1,blank.c];
            aux[blank.l-1,blank.c]:=9;
            new_node := Make_node(pnode, aux, 'UP');
            Enqueue(new_node, plist, enqueue_fn);
         end;
      3: if (blank.l+1<=2) then begin {DOWN}
            Copy_state(pnode^.state, aux);
            aux[blank.l,blank.c]:=aux[blank.l+1,blank.c];
            aux[blank.l+1,blank.c]:=9;
            new_node := Make_node(pnode, aux, 'DOWN');
            Enqueue(new_node, plist, enqueue_fn);
         end;
      end;
   end;
end;

{ ------------------------------------------------------------------------- }
{ funcoes gerais de busca                                                   }
{ ------------------------------------------------------------------------- }

{ funcao de busca geral --------------------------------------------------- }
function General_search(pproblem:problem; enqueue_fn:string):node;
var
   nodes : list;
   tnode, sol : node;
   flag : boolean;

begin
   flag:=true;

   Make_list(nodes, Make_node(nil, pproblem.initial_state,''));

   while (flag) do
   begin
      if IsEmpty(nodes) then
      begin
         flag:=false;
         sol:=nil;
      end
      else begin
         tnode := Remove_front(nodes);

         if (Goal_test(tnode^.state, pproblem.goal_state)) then
         begin
            flag:=false;
            sol:=tnode;
         end
         else begin
            Expand(tnode, nodes, enqueue_fn);
         end;
      end;
   end;

   General_search:=sol;
end;

{ funcao de busca em profundidade limitada -------------------------------- }
function Limited_search(pproblem:problem; enqueue_fn:string; depth:integer):node;
var
   nodes : list;
   tnode, sol : node;
   flag : boolean;

begin
   flag:=true;

   Make_list(nodes, Make_node(nil, pproblem.initial_state,''));

   while (flag) do
   begin
      if IsEmpty(nodes) then
      begin
         flag:=false;
         sol:=nil;
      end
      else begin
         tnode := Remove_front(nodes);

         if (Goal_test(tnode^.state, pproblem.goal_state)) then
         begin
            flag:=false;
            sol:=tnode;
         end
         else if (tnode^.depth < depth) then
            Expand(tnode, nodes, enqueue_fn)
         else {se nao e' solucao nem pode ser expandido => desaloca no}
            Destroy_node(tnode, nodes);  
      end;
   end;

   Limited_search:=sol;
end;


{ funcao de busca em profundidade iterativa ------------------------------- }
function Iterative_search(pproblem:problem; enqueue_fn:string):node;
var
   depth    : integer;
   node_aux : node;

begin
   for depth:=0 to 1000 do
   begin
       nodes_created:=0;

       node_aux := Limited_search(pproblem, enqueue_fn, depth);

       if (node_aux<>nil) then break;  {encontrou solucao}
   end;

   Iterative_search:=node_aux;
end;

{ ------------------------------------------------------------------------- }
{ Metodos nao informados de busca (busca cega)                              }
{ ------------------------------------------------------------------------- }

{ busca em largura}
function Breadth_first_search(pproblem:problem):node;
begin
   Breadth_first_search := General_search(pproblem, 'ENQUEUE_AT_END');
end;

{ busca em profundidade}
function Depth_first_search(pproblem:problem):node;
begin
   Depth_first_search := General_search(pproblem, 'ENQUEUE_AT_FRONT');
end;

{ busca custo uniforme}
function Uniform_cost_search(pproblem:problem):node;
begin
   Uniform_cost_search := General_search(pproblem, 'ENQUEUE_AC_TO_COST');
end;

{ busca limitada em profundidade }
function Depth_limited_search (pproblem:problem; depth:integer):node;
begin
   Depth_limited_search := Limited_search (pproblem,'ENQUEUE_AT_FRONT', depth);
end;

{ busca limitada em profundidade iterativa}
function Iterative_deepening_search (pproblem:problem):node;
begin
   Iterative_deepening_search := Iterative_Search (pproblem,'ENQUEUE_AT_FRONT');
end;

{ ------------------------------------------------------------------------- }
{ Metodos informados de busca                                               }
{ ------------------------------------------------------------------------- }

{ busca gulosa}
function Greedy_search(pproblem:problem):node;
begin
   Greedy_search := General_search(pproblem, 'ENQUEUE_AC_TO_HEUR')
end;

{ busca A*}
function A_search(pproblem:problem):node;
begin
    A_search := General_search(pproblem, 'ENQUEUE_AC_TO_HEUR+COST');
end;

{ executa estrategias de busca -------------------------------------------- }
procedure Execute(var pnode:node);
begin
   if (pnode<>nil) then
      pnode:= Path(pnode)
   else
      writeln('No Solution');

   Destroy(pnode);
end;

{ inicializa jogo_dos_oito -------------------------------------------------}
procedure Initialize(op:integer);
begin
   {estado inicial}
   case op of
   1: begin
         jogo_dos_oito.initial_state[0,0] := 1;
         jogo_dos_oito.initial_state[0,1] := 9;  {    ___________    }
         jogo_dos_oito.initial_state[0,2] := 2;  {   | 1 |   | 2 |   }
         jogo_dos_oito.initial_state[1,0] := 8;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[1,1] := 4;  {   | 8 | 4 | 3 |   }
         jogo_dos_oito.initial_state[1,2] := 3;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[2,0] := 7;  {   | 7 | 6 | 5 |   }
         jogo_dos_oito.initial_state[2,1] := 6;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[2,2] := 5;
      end;
   2: begin
         jogo_dos_oito.initial_state[0,0] := 9;
         jogo_dos_oito.initial_state[0,1] := 2;  {    ___________    }
         jogo_dos_oito.initial_state[0,2] := 3;  {   |   | 2 | 3 |   }
         jogo_dos_oito.initial_state[1,0] := 8;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[1,1] := 1;  {   | 8 | 1 | 4 |   }
         jogo_dos_oito.initial_state[1,2] := 4;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[2,0] := 7;  {   | 7 | 6 | 5 |   }
         jogo_dos_oito.initial_state[2,1] := 6;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[2,2] := 5;
      end;
   3: begin
         jogo_dos_oito.initial_state[0,0] := 8;
         jogo_dos_oito.initial_state[0,1] := 7;  {    ___________    }
         jogo_dos_oito.initial_state[0,2] := 6;  {   | 8 | 7 | 6 |   }
         jogo_dos_oito.initial_state[1,0] := 1;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[1,1] := 9;  {   | 1 |   | 5 |   }
         jogo_dos_oito.initial_state[1,2] := 5;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[2,0] := 2;  {   | 2 | 3 | 4 |   }
         jogo_dos_oito.initial_state[2,1] := 3;  {   |___|___|___|   }
         jogo_dos_oito.initial_state[2,2] := 4;
      end;
   end;

   {estado final}
   jogo_dos_oito.goal_state[0,0] := 1;
   jogo_dos_oito.goal_state[0,1] := 2;  {    ___________    }
   jogo_dos_oito.goal_state[0,2] := 3;  {   | 1 | 2 | 3 |   }
   jogo_dos_oito.goal_state[1,2] := 4;  {   |___|___|___|   }
   jogo_dos_oito.goal_state[2,2] := 5;  {   | 8 |   | 4 |   }
   jogo_dos_oito.goal_state[2,1] := 6;  {   |___|___|___|   }
   jogo_dos_oito.goal_state[2,0] := 7;  {   | 7 | 6 | 5 |   }
   jogo_dos_oito.goal_state[1,0] := 8;  {   |___|___|___|   }
   jogo_dos_oito.goal_state[1,1] := 9;

   {operadores}
   jogo_dos_oito.operators[0] := 'LEFT';
   jogo_dos_oito.operators[1] := 'RIGHT';
   jogo_dos_oito.operators[2] := 'UP';
   jogo_dos_oito.operators[3] := 'DOWN';
end;


{ codigo principal -------------------------------------------------------- }
begin
   Initialize(1);

   clrscr;
   nodes_created:=0;   {numero de nos alocados}
   writeln('Busca em largura');
   writeln;
   problem_node:=Breadth_first_search(jogo_dos_oito);
   Execute(problem_node);
   writeln;
   writeln('Nodes created: ', nodes_created);
   writeln;
   textcolor(17); writeln('Press any key to continue...'); textcolor(7);
   readln;

   {clrscr;
   nodes_created:=0;
   writeln('Busca em profundidade');
   writeln;
   problem_node:=Depth_first_search(jogo_dos_oito);
   Execute(problem_node);
   writeln;
   writeln('Nodes created: ', nodes_created);
   writeln;
   textcolor(17); writeln('Press any key to continue...'); textcolor(7);
   writeln;
   readln;}

   clrscr;
   nodes_created:=0;
   writeln('Busca de custo uniforme');
   writeln;
   problem_node:=Uniform_cost_search(jogo_dos_oito);
   Execute(problem_node);
   writeln;
   writeln('Nodes created: ', nodes_created);
   writeln;
   textcolor(17); writeln('Press any key to continue...'); textcolor(7);
   readln;

   clrscr;
   nodes_created:=0;
   writeln('Busca em profundidade limitada');
   writeln;
   problem_node:=Depth_limited_search(jogo_dos_oito, 10);
   Execute(problem_node);
   writeln;
   writeln('Nodes created: ', nodes_created);
   writeln;
   textcolor(17); writeln('Press any key to continue...'); textcolor(7);
   readln;

   clrscr;
   nodes_created:=0;
   writeln('Busca em profundidade iterativa');
   writeln;
   problem_node:=Iterative_deepening_search(jogo_dos_oito);
   Execute(problem_node);
   writeln;
   writeln('Nodes created: ', nodes_created);
   writeln;
   textcolor(17); writeln('Press any key to continue...'); textcolor(7);
   readln;

   clrscr;
   nodes_created:=0;
   writeln('Busca gulosa');
   writeln;
   problem_node:= Greedy_search(jogo_dos_oito);
   Execute(problem_node);
   writeln;
   writeln('Nodes created: ', nodes_created);
   writeln;
   textcolor(17); writeln('Press any key to continue...'); textcolor(7);
   readln;

   clrscr;
   nodes_created:=0;
   writeln('Busca A*');
   writeln;
   problem_node:= A_search(jogo_dos_oito);
   Execute(problem_node);
   writeln;
   writeln('Nodes created: ', nodes_created);
   writeln;
   textcolor(17); writeln('Press any key to continue...'); textcolor(7);

   repeat until keypressed;
end.
